"""
This module holds the constants used in the other modules.

"""

CACHE_PATH = "cache"
TS_ORDINAL = "ordinal"
TS_TIMESTAMP = "timestamp"

FREQ_D = "d"
FREQ_W = "w"
FREQ_M = "m"
FREQ_Q = "q"
FREQ_Y = "y"

FREQ_H = "h"
FREQ_MIN = "min"
FREQ_SEC = "sec"

FREQ_DAYTYPES = (FREQ_D, FREQ_W, FREQ_M, FREQ_Q, FREQ_Y)
FREQ_IDAYTYPES = (FREQ_H, FREQ_MIN, FREQ_SEC)



"""
This module implements the class Timeseries

"""

from copy import deepcopy
import numpy as np

from .constants import TS_ORDINAL, TS_TIMESTAMP, FREQ_DAYTYPES
from .constants import FREQ_IDAYTYPES, FREQ_D


class TsProto(object):
    """
    This class implements the low-level functions of the Timeseries class.

    By separating the these functions the code base can be broken into more
    managable chunks.

    Most of these functions are handed off the container functions in arrays.
    """

    def __init__(self):

        self.frequency = FREQ_D

        self.tseries = None
        self.dseries = None
        self.end_of_period = True

        self.key = ""
        self.columns = []

    def __iadd__(self, other):
        """ts += other"""
        return self._proc_func(new=False, func="__iadd__", other=other)

    def __irshift__(self, other):
        """'Return self>>=value."""
        return self._proc_func(new=False, func="__irshift__", other=other)

    def __iand__(self, other):
        """Return self&=value."""
        return self._proc_func(new=False, func="__iand__", other=other)

    def __isub__(self, other):
        """ts -= other"""
        return self._proc_func(new=False, func="__isub__", other=other)

    def __ifloordiv__(self, other):
        """ts /= other"""
        return self._proc_func(new=False, func="__ifloordiv__", other=other)

    def __ilshift__(self, other):
        """Return self<<=value."""
        return self._proc_func(new=False, func="__ilshift__", other=other)

    def __itruediv__(self, other):
        """ts /= other"""
        return self._proc_func(new=False, func="__itruediv__", other=other)

    def __imod__(self, other):
        """ts %= other"""
        return self._proc_func(new=False, func="__imod__", other=other)

    def __ior__(self, other):
        """Return self|=value."""
        return self._proc_func(new=False, func="__ior__", other=other)

    def __ixor__(self, other):
        """Return self^=value."""
        return self._proc_func(new=False, func="__ixor__", other=other)

    def __imul__(self, other):
        """ts *= other"""
        return self._proc_func(new=False, func="__imul__", other=other)

    def __ipow__(self, other):
        """ts **= other"""
        return self._proc_func(new=False, func="__ipow__", other=other)

    def __radd__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__radd__", other=other)

    def __rtruediv__(self, other):
        """Return other / ts."""
        return self._proc_func(new=True, func="__rtruediv__", other=other)

    def __rand__(self, other):
        """Return other & ts."""
        return self._proc_func(new=True, func="__rand__", other=other)

    def __rxor__(self, other):
        """Return value^self."""
        return self._proc_func(new=True, func="__rxor__", other=other)

    def __rdivmod__(self, other):
        """Return divmod(value, self)"""
        return self._proc_func(new=True, func="__rdivmod__", other=other)

    def __rmod__(self, other):
        """Return other % ts."""
        return self._proc_func(new=True, func="__rmod__", other=other)

    def __rmul__(self, other):
        """Return other * ts."""
        return self._proc_func(new=True, func="__rmul__", other=other)

    def __rfloordiv__(self, other):
        """Return value//self."""
        return self._proc_func(new=True, func="__rfloordiv__", other=other)

    def __rlshift__(self, other):
        """Return self<<value."""
        return self._proc_func(new=True, func="__rlshift__", other=other)

    def __rsub__(self, other):
        """Return other - ts."""
        return self._proc_func(new=True, func="__rsub__", other=other)

    def __ror__(self, other):
        """Return value|self"""
        return self._proc_func(new=True, func="__ror__", other=other)

    def __rpow__(self, other):
        """Return other ** ts."""
        return self._proc_func(new=True, func="__rpow__", other=other)

    def __rrshift__(self, other):
        """Return value>>self."""
        return self._proc_func(new=True, func="__rrshift__", other=other)

    def __abs__(self):
        """Return abs(ts)."""
        return self._proc_func(new=True, func="__abs__", other=None)

    def __pos__(self):
        """Return +ts."""
        return self._proc_func(new=True, func="__pos__", other=None)

    def __neg__(self):
        """Return -ts."""
        return self._proc_func(new=True, func="__neg__", other=None)

    def __invert__(self):
        """not ts."""
        return self._proc_func(new=True, func="__invert__", other=None)

    def __add__(self, other):
        """ts + other.

        Related functions:
            add()
        """
        return self._proc_func(new=True, func="__add__", other=other)

    def __pow__(self, other):
        """Return ts ** other."""
        return self._proc_func(new=True, func="__pow__", other=other)

    def __and__(self, other):
        """Return ts & other."""
        return self._proc_func(new=True, func="__and__", other=other)

    def __divmod__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__divmod__", other=other)

    def __eq__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__eq__", other=other)

    def __floordiv__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__floordiv__", other=other)

    def __ge__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__ge__", other=other)

    def __sub__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__sub__", other=other)

    def __truediv__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__truediv__", other=other)

    def __gt__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__gt__", other=other)

    def __xor__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__xor__", other=other)

    def __le__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__le__", other=other)

    def __rshift__(self, other):
        """Return self>>value."""
        return self._proc_func(new=True, func="__rshift__", other=other)

    def __lshift__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__lshift__", other=other)

    def __lt__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__lt__", other=other)

    def __mod__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__mod__", other=other)

    def __mul__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__mul__", other=other)

    def __ne__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__ne__", other=other)

    def __or__(self, other):
        """Return other + ts."""
        return self._proc_func(new=True, func="__or__", other=other)

    def _proc_func(self, new, func, other=None):
        """Processes container function.

        Reorganize this.
        """

        if new:
            tmp_ts = self.clone()

            if isinstance(other, TsProto):
                self._column_checks(tmp_ts, other)
                tmp_ts, other = self.common_length(tmp_ts, other)

            if other is None:
                # there may be others that should be in this list.
                #   at the moment, fixing a problem.
                if func in ("__eq__", "__ne__"):
                    tmp_ts.tseries = getattr(tmp_ts.tseries, func)(other)
                else:
                    tmp_ts.tseries = getattr(tmp_ts.tseries, func)()
            else:
                if isinstance(other, TsProto):
                    tmp_ts.tseries = getattr(tmp_ts.tseries, func)(
                        other.tseries
                    )
                else:
                    tmp_ts.tseries = getattr(tmp_ts.tseries, func)(other)

            return tmp_ts
        else:
            if other is None:
                self.tseries = (getattr(self.tseries, func)(),)
                return self
            else:
                if isinstance(other, TsProto):
                    self._column_checks(self, other)
                    self, other = self.common_length(self, other)
                    self.tseries = getattr(self.tseries, func)(other.tseries)
                else:
                    self.tseries = getattr(self.tseries, func)(other)
                return self

    @staticmethod
    def _column_checks(ts1, ts2):
        """
        This function raises an error if the columns on the two timeseries
        do not match.
        """
        status = False
        if len(ts1.tseries.shape) == 2 and len(ts2.tseries.shape) != 2:
            status = True
        elif len(ts1.tseries.shape) != 2 and len(ts2.tseries.shape) == 2:
            status = True

        elif len(ts1.tseries.shape) == 2 and len(ts2.tseries.shape) == 2:
            if ts1.tseries.shape[1] != ts2.tseries.shape[1]:
                status = True

        else:
            pass

        if status:
            raise ValueError(
                " ".join(
                    ["Both timeseries must have the same columns:", "%s vs %s"]
                )
                % (ts1.tseries.shape, ts2.tseries.shape)
            )

    @staticmethod
    def _array_size_check(ts1, ts2):
        """Raise error if timeseries' shapes do not match."""
        if np.array_equal(ts1, ts2):
            return True
        else:
            raise ValueError(
                " ".join(
                    ["Sizes of the timeseries are different:", "%s vs %s"]
                )
                % (ts1.tseries.shape, ts2.tseries.shape)
            )

    def make_arrays(self):
        """
        Convert the date and time series lists (if so) to numpy arrays
        """
        self.tseries = self._make_array(self.tseries, np.float64)

        if self.get_date_series_type() == TS_ORDINAL:
            self.dseries = self._make_array(self.dseries, np.int32).flatten()
        else:
            self.dseries = self._make_array(self.dseries, np.float64).flatten()

    @staticmethod
    def _make_array(convert_list, numtype):
        """
        Converts a list to numpy array
        """
        return np.array(convert_list, numtype)

    def lengths(self):
        """
        This function returns the lengths of both the date series and time
        series. Both numbers are included in case a mismatch has occurred.
        """
        return (len(self.dseries), len(self.tseries))

    def get_date_series_type(self):
        """
        This function returns the date series type associated with the
        timeseries.  The choices are TS_ORDINAL or TS_TIMESTAMP.

        """

        if self.frequency in FREQ_DAYTYPES:
            return TS_ORDINAL
        elif self.frequency in FREQ_IDAYTYPES:
            return TS_TIMESTAMP
        else:
            raise ValueError("Unknown frequency: %s" % self.frequency)

    def __getitem__(self, key):
        """
        This function returns a timeseries where both the date and values are
        sliced.

        Usage:
            ts[:10]

            returns a timeseries with dseries and tseries of length 10.

        """

        if isinstance(key, tuple):
            dkey = key[0]
        else:
            dkey = key

        tmp_ts = self.clone()
        tmp_ts.dseries = tmp_ts.dseries[dkey]
        tmp_ts.tseries = tmp_ts.tseries[key]

        return tmp_ts

    def clone(self):
        """This function returns a copy of the timeseries."""
        return deepcopy(self)

    @staticmethod
    def common_length(*ts):
        """
        This function trims the lengths of timeseries and returns all
        timeseries with the same length.

        The idea is that in order to do array operations there must be a
        common length for each timeseries.

        Reflecting the bias for using timeseries sorted from latest info to
        earlier info, truncation takes place at the end of the array. That
        way older less important values are removed if necessary.

        This function does not alter the timeseries passed in. The list of
        timeseries returned are clones of the originals.

        Changed:

        Usage:
            [ts1, ts2, ..., ts_n] = self.common_length(*ts)

        Formerly:
            ts1_new, ts2_new = self.common_length(ts1, ts2)


        """

        min_length = min([len(ts_tmp.tseries) for ts_tmp in ts])

        return [ts_tmp[:min_length] for ts_tmp in ts]

    def shape(self):
        """
        This function return the shape of the timeseries. This is a shortcut
        to putting in ts.tseries.shape.
        """

        if self.tseries is None:
            return None
        elif isinstance(self.tseries, list):
            return np.array(self.tseries).shape
        else:
            return self.tseries.shape

    def if_dseries_match(self, ts):
        """
        This function returns True if the date series are the same.
        """

        return np.array_equal(self.dseries, ts.dseries)

    def if_tseries_match(self, ts):
        """
        This function returns True if the time series are the same.
        """

        return np.array_equal(self.tseries, ts.tseries)



# freq_conversions.py
"""
This module converts timeseries from one type to another.

"""

import numpy as np

from .constants import FREQ_D, FREQ_M, FREQ_Q, FREQ_Y
from .constants import FREQ_H, FREQ_MIN, FREQ_SEC
from .constants import TS_ORDINAL, TS_TIMESTAMP

HIERARCHY = (FREQ_SEC, FREQ_MIN, FREQ_H, FREQ_D, FREQ_M, FREQ_Q, FREQ_Y)


def _q_test(date, kwargs):
    """
    Computes quarterly indicators.

    No expectation of using kwargs in this. It is for consistency.
    """

    return date.month % 3 == 0


def _weekday_test(date, kwargs):
    """
    Computes weekly indicators.
    """

    if "weekday" in kwargs:
        weekday = kwargs["weekday"]

        if date.weekday() == weekday:
            return 1
        else:
            return 0
    else:
        return date.weekday()


def _filter_dates(dates, freq, kwargs):
    """
    This function filters dates to indicate end of periods for ordinals.
    """

    indicator = DATETIME_DICT[freq]

    if isinstance(indicator, str):
        # no special behavior
        indicators = np.fromiter(
            [date.__getattribute__(indicator) for date in dates],
            dtype=np.int32,
        )

        return np.argwhere(indicators[1:] - indicators[:-1] > 0)

    else:
        # apply a function
        indicators = np.fromiter(
            [indicator(date, kwargs) for date in dates], dtype=np.int32
        )

        return np.argwhere(indicators[1:] - indicators[:-1] > 0)


def _filter_idates(dates, freq, end_of_period, **kwargs):
    """
    This function filters dates to indicate end of periods for timestamps.
    """

    indicator = DATETIME_DICT[freq]

    if isinstance(indicator, str):
        # no special behavior
        indicators = np.fromiter(
            [date.__getattribute__(indicator) for date in dates],
            dtype=np.int32,
        )

        selected = np.argwhere(indicators[1:] - indicators[:-1] > 0)

        # check special case of start date

        if end_of_period is False:
            if dates[-1].__getattribute__(indicator) == 0:
                selected = np.append(selected, len(dates) - 1)

        return selected

    else:
        # apply a function -- here for completeness at the moment
        # could apply to 5 minute data for example
        return np.fromiter(
            [indicator(date, kwargs) for date in dates], dtype=np.int32
        )


DATETIME_DICT = {
    # 'sec': 'second',
    "min": "second",
    "h": "minute",
    "d": "hour",
    "w": _weekday_test,
    "m": "day",
    "q": _q_test,
    "y": "month",
}


def convert(ts, new_freq, include_partial=True, **kwargs):
    """
    This function converts a timeseries to another frequency. Conversion only
    works from a higher frequency to a lower frequency, for example daily to
    monthly.

    NOTE: add a gatekeeper for invalid kwargs.
    """

    new_ts = ts.clone()
    series_dir = ts.series_direction()
    new_ts.sort_by_date(reverse=True)

    freq_idx = HIERARCHY.index(ts.frequency)
    new_idx = HIERARCHY.index(FREQ_Q)
    daily_idx = HIERARCHY.index(FREQ_D)

    if freq_idx > new_idx:

        raise ValueError(
            "Cannot convert from %s to %s." % (ts.frequency, new_freq)
        )

    dates = new_ts.datetime_series()

    date_series_type = ts.get_date_series_type()
    if date_series_type == TS_ORDINAL:
        selected = _filter_dates(dates, new_freq, kwargs)
    elif date_series_type == TS_TIMESTAMP:
        selected = _filter_idates(
            dates, new_freq, end_of_period=ts.end_of_period
        )
    else:
        raise ValueError("Invalid date series type: %s" % (date_series_type))

    if selected.shape[0] > 0:

        if new_ts.end_of_period:
            selected += 1  # shift to start of next period

        if include_partial or freq_idx > daily_idx:
            if selected[0] != 0:
                # insert most recent date
                # selected = np.insert(selected, 0, 0)
                # np.insert(arr, obj, values, axis=None)
                selected = np.insert(selected, 0, 0)

        if freq_idx > daily_idx:
            # already processed (probably)
            if selected[-1] != len(dates) - 1:
                selected = np.append(selected, len(dates) - 1)

    new_ts.tseries = new_ts.tseries[selected.flatten()]

    new_ts.frequency = new_freq

    if new_freq == FREQ_D:
        # convert dates from timestamp to ordinal
        new_ts.dseries = np.fromiter(
            [date.toordinal() for date in np.array(dates)[selected]],
            dtype=np.int32,
        )
    else:
        new_ts.dseries = new_ts.dseries[selected]

    new_ts.dseries = new_ts.dseries.flatten()

    if series_dir != new_ts.series_direction():
        new_ts.reverse()

    return new_ts



"""
This module implements a class derived from lists for working with timeseries.

"""

from copy import deepcopy
import json

from .timeseries import Timeseries


class TssList(list):
    """
    This class is a way of handling some the routine tasks for groups
    of timeseries.

    Assumption:
        This is a list of timeseries of common interest

    Usage:
        tss = TssList(tss=None)  # where tss is a list of timeseries


    """

    def __init__(self, tss=None):

        list.__init__(self)  # only did this to satisfy pylint
        if tss is None:
            pass
        elif isinstance(tss, list):
            self.extend(tss)
        else:
            raise ValueError("Initialization requires either a list or None")

    def min_date(self):
        """
        Returns the earliest date as a datetime.

        By using a datetime, there is no need to distinguish whether time
        series are ordinal or timestamp or mixed.

        """

        dates = [
            ts.start_date("datetime") for ts in self if ts.dseries is not None
        ]

        if dates:
            return min(dates)
        else:
            return None

    def max_date(self):
        """
        Returns the latest date

        """

        dates = [
            ts.end_date("datetime") for ts in self if ts.dseries is not None
        ]

        if dates:
            return max(dates)
        else:
            return None

    def get_values(self, date, notify=False):
        """
        This function finds the values current to the date for the tickers.

        The point of notify is not fail gracefully if necessary.
        """

        values = []
        for idx, ts_tmp in enumerate(self):

            try:
                values.append(ts_tmp.tseries[ts_tmp.row_no(date)])
            except ValueError:
                if notify:
                    raise ValueError(
                        "ts %s does not have a value on %s" % (idx, date)
                    )
                else:
                    values.append(None)

        return tuple(values)

    def combine(self, discard=True, pad=None):
        """
        This function combines all timeseries into one by applying the combine
        function to each timeseries.

        Usage:
            combine(self, discard=True, pad=None)

        returns ts

        """

        if len(self) > 1:
            return self[0].combine(self[1:], discard=discard, pad=pad)
        else:
            return self[0].clone()

    def clone(self):
        """
        Returns a new copy of the object.
        """

        return deepcopy(self)

    def as_dict(self):
        """
        This function returns a dict with keys being the timeseries key.

        If keys are missing, an error is raised.
        """

        for ts_tmp in self:
            if ts_tmp.key == "" or ts_tmp.key is None:
                raise ValueError("There must be a key for each timeseries.")

        return dict([(ts_tmp.key, ts_tmp) for ts_tmp in self])

    def to_list(self, dt_fmt="str", data_list=True):
        """
        This function builds a list of timeseries that are in the form of
        dict objects for use when build JSON formatted file.
        """

        outlist = []

        for ts in self:
            outlist.append(ts.to_dict(dt_fmt=dt_fmt, data_list=data_list))

        return outlist

    def to_json(self, indent=2, dt_fmt="str", data_list=True):
        """
        This function returns the timeseries list in JSON format.

        Usage:
            self.to_json(indent=2, dt_fmt='str', data_list=True)

        indent: indenting in the JSON output
        dt_fmt: formatting of the dates. Look at help for
                    Timeseries.to_dict
        data_list: Whether data uses a dict for dates as keys or
                   simply a list.
                   Default is for a list. Otherwise, sorting the
                   timeseries in the list would be required.

        """
        return json.dumps(
            self.to_list(dt_fmt=dt_fmt, data_list=data_list), indent=indent
        )

    def from_json(self, json_str):
        """
        This function loads a JSON string and applies it to the object.
        """
        self.clear()

        tss_tmp = json.loads(json_str)

        for item in tss_tmp:
            self.append(Timeseries().from_dict(item))

        return self



"""
This module tests the prototype timeseries class.
"""
import unittest

from datetime import datetime
import numpy as np

from thymus.constants import FREQ_D, FREQ_SEC
from thymus.tsproto import TsProto
from thymus.tsslist import TssList


class TestTsProto(unittest.TestCase):
    """
    This class tests the base class TsProto.
    """

    def setUp(self):

        # three timeseries
        self.ts = TsProto()
        start_date = datetime(2015, 12, 31).toordinal()
        self.ts.dseries = start_date + np.arange(10)
        self.ts.tseries = np.arange(10)
        self.ts.make_arrays()

        # longer timeseries
        self.ts_long = TsProto()
        start_date = datetime(2015, 12, 31).toordinal()
        self.ts_long.dseries = start_date + np.arange(20)
        self.ts_long.tseries = np.arange(20)
        self.ts_long.make_arrays()

        # shorter timeseries
        self.ts_short = TsProto()
        start_date = datetime(2015, 12, 31).toordinal()
        self.ts_short.dseries = start_date + np.arange(5)
        self.ts_short.tseries = np.arange(5)
        self.ts_short.make_arrays()

    def test_class_init_(self):
        """Test class initialization."""
        tmp_ts = TsProto()

        # defaults for dseries, tseries might get changed back to list
        # self.assertIsInstance(tmp_ts.dseries, list)
        # self.assertIsInstance(tmp_ts.tseries, list)
        self.assertIsInstance(tmp_ts.columns, list)
        self.assertEqual(tmp_ts.frequency, FREQ_D)
        self.assertTrue(tmp_ts.end_of_period)

    def test_container_functions(self):
        """Tests the ability to pass through container functions to array."""

        ts = self.ts.clone()

        flist = [
            "__pow__",
            "__add__",
            "__rsub__",
            "__rtruediv__",
            "__divmod__",
            "__sub__",
            "__eq__",
            "__floordiv__",
            "__truediv__",
            "__ge__",
            "__gt__",
            "__le__",
            "__lt__",
            "__mod__",
            "__mul__",
            "__ne__",
            "__radd__",
            "__rdivmod__",
            "__rfloordiv__",
            "__rmod__",
            "__rmul__",
            "__rpow__",
        ]

        unary_flist = ["__abs__", "__pos__", "__neg__"]

        special_flist = ["__invert__"]

        flist1 = [
            "__and__",
            "__or__",
            "__ror__",
            "__rand__",
            "__rxor__",
            "__xor__",
            "__rshift__",
            "__rlshift__",
            "__lshift__",
            "__rrshift__",
        ]

        iflist = [
            "__iadd__",
            "__ifloordiv__",
            "__imod__",
            "__imul__",
            "__ipow__",
            "__isub__",
            "__itruediv__",
        ]

        iflist1 = [
            "__iand__",
            "__ilshift__",
            "__ior__",
            "__irshift__",
            "__ixor__",
        ]

        ts_other = self.ts.clone() * 4 + 3
        for func in flist:
            for other in [3, ts_other]:

                ts = self.ts.clone()
                ts.tseries += 1

                ts_a = getattr(ts, func)(other)
                a_series = ts_a.tseries

                if isinstance(other, TsProto):
                    b_series = getattr(ts.tseries, func)(other.tseries)
                else:
                    b_series = getattr(ts.tseries, func)(other)

                self.assertTrue(np.array_equal(a_series, b_series))
                self.assertTrue(ts.if_dseries_match(ts_a))

        # specific tests
        # decide what to do here. future warning that it will
        #    be eventually elementwise
        self.assertIsNotNone(ts_other)

        for func in unary_flist:
            ts = self.ts.clone()
            ts_a = getattr(ts, func)()
            a_series = ts_a.tseries

            b_series = getattr(ts.tseries, func)()

            self.assertTrue(np.array_equal(a_series, b_series))
            self.assertTrue(ts.if_dseries_match(ts_a))

        for func in special_flist:
            ts = self.ts.clone()
            ts.tseries = np.array(ts.tseries, np.int32)
            self.assertTrue(
                np.array_equal(
                    getattr(ts, func)().tseries, getattr(ts.tseries, func)()
                )
            )

        ts_other = self.ts.clone() * 4
        ts_other.tseries = np.array(ts_other.tseries, np.int64)
        for func in flist1:
            for other in [3, ts_other]:
                ts = self.ts.clone()
                ts.tseries = np.array(ts.tseries, np.int64)

                ts_a = getattr(ts, func)(other)
                a_series = ts_a.tseries

                if isinstance(other, TsProto):
                    b_series = getattr(ts.tseries, func)(other.tseries)
                else:
                    b_series = getattr(ts.tseries, func)(other)

                self.assertTrue(np.array_equal(a_series, b_series))
                self.assertTrue(ts.if_dseries_match(ts_a))

        for func in iflist:
            for other in [3, self.ts.clone() + 2.5]:
                ts = self.ts.clone() + 4.0
                ts1 = self.ts.clone() + 4.0

                ts_a = getattr(ts, func)(other)
                a_series = ts_a.tseries

                if isinstance(other, TsProto):
                    b_series = getattr(ts.tseries, func)(other.tseries)
                else:
                    b_series = getattr(ts.tseries, func)(other)

                self.assertTrue(np.array_equal(a_series, b_series))
                self.assertTrue(ts.if_dseries_match(ts_a))

        ts_other = self.ts.clone() * 4
        ts_other.tseries = np.array(ts_other.tseries, np.int64)
        for func in iflist1:
            for other in [3, ts_other]:
                ts = self.ts.clone()
                ts.tseries = np.array(ts.tseries, np.int64)
                ts1 = self.ts.clone()
                ts1.tseries = np.array(ts1.tseries, np.int64)

                if other is None:
                    other = self.ts.clone() * 4
                    other.tseries = np.array(other.tseries, np.int64)

                ts_a = getattr(ts, func)(other)
                a_series = ts_a.tseries

                if isinstance(other, TsProto):
                    b_series = getattr(ts.tseries, func)(other.tseries)
                else:
                    b_series = getattr(ts.tseries, func)(other)

                self.assertTrue(np.array_equal(a_series, b_series))
                self.assertTrue(ts.if_dseries_match(ts_a))

    def test_timeseries__add__lengths(self):
        """Tests adding two timeseries with mismatched lengths"""
        #
        # silently truncate
        ts1 = self.ts.clone()
        ts1.dseries = self.ts.dseries.tolist()[:5]
        ts1.tseries = self.ts.tseries.tolist()[:5]
        ts1.make_arrays()

        ts = self.ts + ts1

        self.assertEqual(ts.tseries[0], 0)
        self.assertEqual(ts.tseries[1], 2)
        self.assertEqual(ts.tseries[2], 4)
        self.assertEqual(ts.tseries[3], 6)
        self.assertEqual(ts.tseries[4], 8)

        self.assertEqual(len(ts.dseries), 5)
        self.assertEqual(len(ts.tseries), 5)

    def test_timeseries__add__columns(self):
        """Tests adding two timeseries mismatched columns"""
        ts1 = self.ts.clone()
        ts1.tseries = np.arange(10).reshape((-1, 1))
        ts1.make_arrays()

        # self ts shape   (10) incoming (10, 1)
        self.assertRaises(ValueError, self.ts.__add__, ts1)

        # self ts shape   (10, 1) incoming (10, 2)
        ts2 = ts1.clone()
        ts2.tseries = np.arange(20).reshape((-1, 2))
        ts2.make_arrays()

        self.assertRaises(ValueError, ts1.__add__, ts2)

    def test_timeseries__iadd__lengths(self):
        """Tests in-place adding two timeseries mismatched lengths"""
        #
        # silently truncate
        ts1 = self.ts.clone()
        ts1.dseries = self.ts.dseries.tolist()[:5]
        ts1.tseries = self.ts.tseries.tolist()[:5]
        ts1.make_arrays()

        self.ts += ts1

        self.assertEqual(self.ts.tseries[0], 0)
        self.assertEqual(self.ts.tseries[1], 2)
        self.assertEqual(self.ts.tseries[2], 4)
        self.assertEqual(self.ts.tseries[3], 6)
        self.assertEqual(self.ts.tseries[4], 8)

        self.assertEqual(len(self.ts.dseries), 5)
        self.assertEqual(len(self.ts.tseries), 5)

    def test_timeseries__iadd__columns(self):
        """Tests in-place adding two timeseries with mismatched columns"""
        ts1 = self.ts.clone()
        ts1.tseries = np.arange(10).reshape((-1, 1))
        ts1.make_arrays()

        # self ts shape   (10) incoming (10, 1)
        self.assertRaises(ValueError, self.ts.__iadd__, ts1)

        # self ts shape   (10, 1) incoming (10, 2)
        ts2 = ts1.clone()
        ts2.tseries = np.arange(20).reshape((-1, 2))
        ts2.make_arrays()

        self.assertRaises(ValueError, ts1.__iadd__, ts2)

    def test_column_checks(self):
        """Tests verification of similar columns """
        ts1 = self.ts.clone()
        ts1.tseries = np.arange(10).reshape((-1, 1))
        ts1.make_arrays()

        # self ts shape   (10) incoming (10, 1)
        self.assertRaises(ValueError, self.ts._column_checks, self.ts, ts1)

        # self ts shape   (10, 1) incoming (10, 2)
        ts2 = ts1.clone()
        ts2.tseries = np.arange(20).reshape((-1, 2))
        ts2.make_arrays()

        self.assertRaises(ValueError, self.ts._column_checks, ts1, ts2)

        # now one with columns ok
        ts2 = ts1.clone()

        self.assertIsNone(self.ts._column_checks(ts1, ts2))

    def test_if_dseries_match(self):
        """Tests comparing two date series."""

        ts = self.ts.clone()

        self.assertTrue(self.ts.if_dseries_match(ts))

        ts = self.ts.clone()
        ts.dseries[0] += 1

        self.assertFalse(self.ts.if_dseries_match(ts))

    def test_if_tseries_match(self):
        """Tests comparing two series of values."""

        ts = self.ts.clone()

        self.assertTrue(self.ts.if_tseries_match(ts))

        ts = self.ts.clone()
        ts.tseries[0] += 1

        self.assertFalse(self.ts.if_tseries_match(ts))

    def test___getitem__(self):
        """This function tests selection."""

        ts = self.ts.clone()

        ts1 = ts[:2]

        self.assertTrue(np.array_equal(ts1.dseries, ts.dseries[:2]))
        self.assertTrue(np.array_equal(ts1.tseries, ts.tseries[:2]))

        # test separate slicing for dseries and tseries
        ts1 = ts.clone()

        ts1.tseries = np.arange(len(ts.tseries) * 4).reshape((-1, 4))

        ts2 = ts1[:5, 1]
        self.assertTrue(np.array_equal(ts2.dseries, ts1.dseries[:5]))
        self.assertTrue(np.array_equal(ts2.tseries, ts1.tseries[:5, 1]))

        # test separate slicing with more dimensions in tseries
        ts1 = TsProto()
        ts1.dseries = datetime(2016, 1, 1).toordinal() + np.arange(1000)

        ts1.tseries = np.arange(9000).reshape((1000, 3, 3))

        ts2 = ts1[:500, 1]

        self.assertTrue(np.array_equal(ts2.dseries, ts1.dseries[:500]))
        self.assertTrue(np.array_equal(ts2.tseries, ts1.tseries[:500, 1]))

        ts2 = ts1[:500, :, 1]

        self.assertTrue(np.array_equal(ts2.dseries, ts1.dseries[:500]))
        self.assertTrue(np.array_equal(ts2.tseries, ts1.tseries[:500, :, 1]))

        ts2 = ts1[:500, 1, :2]

        self.assertTrue(np.array_equal(ts2.dseries, ts1.dseries[:500]))
        self.assertTrue(np.array_equal(ts2.tseries, ts1.tseries[:500, 1, :2]))

    def test_timeseries_common_length(self):
        """Tests truncating timeseries to a common length."""

        ts1, ts2 = self.ts.common_length(self.ts, self.ts_long)
        self.assertEqual(len(ts1.tseries), len(ts2.tseries))

        ts1, ts2, ts3 = self.ts.common_length(
            self.ts, self.ts_short, self.ts_long
        )
        self.assertEqual(len(ts1.tseries), len(ts2.tseries))
        self.assertEqual(len(ts1.tseries), len(ts3.tseries))

    def test_timeseries_shape(self):
        """Tests returning the shape of the tseries."""

        self.assertTupleEqual(self.ts.shape(), self.ts.tseries.shape)

        ts = TsProto()

        # test blank TsProto
        ts = TsProto()
        self.assertIsNone(ts.shape(), None)

        # now with data
        ts.dseries = np.arange(100)
        ts.tseries = np.arange(200).reshape(100, 2)

        self.assertTupleEqual(ts.shape(), (ts.tseries.shape))

    def test_lengths(self):
        """Tests returning the lengths of the dseries and tseries."""

        lengths = self.ts.lengths()

        self.assertEqual(lengths[0], len(self.ts.dseries))
        self.assertEqual(lengths[1], len(self.ts.tseries))

        self.ts.tseries = self.ts.tseries[:3]

        lengths = self.ts.lengths()

        self.assertEqual(lengths[0], len(self.ts.dseries))
        self.assertEqual(lengths[1], len(self.ts.tseries))

    def test_timeseries_make_arrays(self):
        """Tests converting lists to arrays for both dseries and tseries."""

        # daily, so ordinal
        ts = TsProto()

        ts.frequency = FREQ_D

        ts.dseries = [i for i in range(100)]
        ts.tseries = [i for i in range(100)]

        ts.make_arrays()

        self.assertTrue(np.array_equal(ts.dseries, np.arange(100)))

        self.assertTrue(isinstance(ts.dseries[0], np.int32))
        self.assertTrue(isinstance(ts.tseries[0], np.float64))

        # seconds, so timestamp
        ts = TsProto()

        ts.frequency = FREQ_SEC

        ts.dseries = [i for i in range(100)]
        ts.tseries = [i for i in range(100)]

        ts.make_arrays()

        self.assertTrue(np.array_equal(ts.dseries, np.arange(100)))

        self.assertTrue(isinstance(ts.dseries[0], np.float64))
        self.assertTrue(isinstance(ts.tseries[0], np.float64))

        # verify that dseries is flattened
        ts.dseries = [[i] for i in range(100)]
        ts.tseries = [i for i in range(100)]

        ts.make_arrays()
        self.assertEqual(len(ts.dseries.shape), 1)

    def test_timeseries__make_array(self):
        """Tests making a numpy array to a specific type."""

        ts = TsProto()

        convert_list = [i for i in range(100)]

        new_array = ts._make_array(convert_list, numtype=np.float64)

        # verify structure, does not verify type
        self.assertTrue(np.array_equal(new_array, np.array(convert_list)))

        self.assertTrue(isinstance(new_array[0], np.float64))

        new_array = ts._make_array(convert_list, numtype=np.int32)

        self.assertTrue(np.array_equal(new_array, np.array(convert_list)))

        self.assertTrue(isinstance(new_array[0], np.int32))

    def test_timeseries_clone(self):
        """Tests creation of duplication timeseries."""

        ts = self.ts.clone()

        # is it a separate object
        self.assertNotEqual(ts.__str__(), self.ts.__str__())

        # do the characteristics match up?
        self.assertEqual(ts.key, self.ts.key)
        self.assertEqual(ts.frequency, self.ts.frequency)
        self.assertTrue(np.array_equal(ts.tseries, self.ts.tseries))
        self.assertTrue(np.array_equal(ts.dseries, self.ts.dseries))
        self.assertListEqual(ts.columns, self.ts.columns)
        self.assertEqual(ts.end_of_period, self.ts.end_of_period)


if __name__ == "__main__":
    unittest.main()



"""
This module tests the conversions from one frequency to another.

"""

from datetime import datetime, timedelta
import numpy as np

import unittest

from thymus.timeseries import Timeseries
from thymus.constants import FREQ_D, FREQ_W, FREQ_M, FREQ_Q, FREQ_Y
from thymus.constants import FREQ_H, FREQ_MIN

from thymus.freq_conversions import convert


class TestFreqConversions(unittest.TestCase):
    """
    This class tests conversions of timeseries.
    """

    def setUp(self):

        # sample timeseries
        self.ts_ord = Timeseries()
        start_date = datetime(2015, 12, 31)

        # sloppily ends slightly more than two years
        end_date = datetime(2018, 1, 15)

        # set up two years of data with weekends skipped
        date = start_date
        self.ts_ord.dseries = []
        while date <= end_date:
            if date.weekday() not in [5, 6]:
                self.ts_ord.dseries.append(date.toordinal())
            date += timedelta(days=1)

        self.ts_ord.tseries = np.arange(len(self.ts_ord.dseries))
        self.ts_ord.make_arrays()

        # timestamp based timeseries
        self.ts_seconds = Timeseries(frequency="sec")
        start_date = datetime(2016, 1, 1, 0, 0)
        end_date = datetime(2016, 1, 4, 0, 0)

        length = (end_date - start_date).total_seconds()

        self.ts_seconds.dseries = start_date.timestamp() + np.arange(length)
        self.ts_seconds.tseries = np.arange(length)
        self.ts_seconds.make_arrays()

    def test_convweekly_period_start(self):
        """ Test timeseries conversion to weekly with start-of-period data. """
        ts = self.ts_ord.clone()
        ts.end_of_period = False

        # conv_weekly with defaults
        ts1 = convert(ts, new_freq=FREQ_W)

        self.assertEqual(ts1.frequency, FREQ_W)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 4).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 11).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 18).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 25).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 2, 1).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 8).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 8).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_weekly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_W, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 4).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 11).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 18).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 25).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 2, 1).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 8).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 8).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        ts1 = convert(ts, new_freq=FREQ_W, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 4).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 11).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 18).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 25).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 2, 1).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 8).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        ts1 = convert(ts, new_freq=FREQ_W, weekday=2)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 7).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 14).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 21).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 28).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 2, 4).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 11).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 11).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # test lower frequency data
        self.assertRaises(
            ValueError,
            convert,
            convert(self.ts_ord, new_freq=FREQ_Y),
            new_freq=FREQ_W,
        )

    def test_convweekly_period_end(self):
        """
        Test timeseries conversion to weekly with end-of-period data.
        """
        ts = self.ts_ord.clone()

        # conv_weekly with defaults
        ts1 = convert(ts, new_freq=FREQ_W)

        self.assertEqual(ts1.frequency, FREQ_W)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 8).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 15).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 22).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 1, 29).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 5).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 12).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_weekly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_W, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 8).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 15).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 22).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 1, 29).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 5).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 12).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_weekly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_W, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 8).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 15).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 22).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 1, 29).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 5).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 12).toordinal())

        # conv_weekly with weekday=2
        ts1 = convert(ts, new_freq=FREQ_W, weekday=2)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 6).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 13).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 1, 20).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 1, 27).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 2, 3).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 2, 10).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 10).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # test lower frequency data
        self.assertRaises(
            ValueError,
            convert,
            convert(self.ts_ord, new_freq=FREQ_Y),
            new_freq=FREQ_W,
        )

    def test_convmonthly_period_start(self):
        """
        This function tests converting timeseries to monthly data that have
        starting period data.

        """

        # daily_monthly(ts, new_freq):
        ts = self.ts_ord.clone()
        ts.end_of_period = False

        # daily_monthly with defaults
        ts1 = convert(ts, new_freq=FREQ_M)

        self.assertEqual(ts1.frequency, FREQ_M)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 2, 1).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 3, 1).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 4, 1).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 5, 2).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 6, 1).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2016, 7, 1).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_monthly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_M, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 2, 1).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 3, 1).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 4, 1).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 5, 2).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 6, 1).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2016, 7, 1).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_monthly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_M, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 2, 1).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 3, 1).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 4, 1).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 5, 2).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 6, 1).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2016, 7, 1).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 1).toordinal())

        # test lower frequency data
        self.assertRaises(
            ValueError,
            convert,
            convert(self.ts_ord, new_freq="y"),
            new_freq=FREQ_M,
        )

        # timestamp conversion goes here

    def test_convmonthly_period_end(self):
        """
        This function tests converting timeseries to monthly data that have
        ending period data.

        """
        ts = self.ts_ord.clone()

        # daily_monthly with defaults
        ts1 = convert(ts, new_freq=FREQ_M)

        self.assertEqual(ts1.frequency, FREQ_M)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 29).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 2, 29).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 4, 29).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 5, 31).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2016, 6, 30).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2017, 12, 29).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_monthly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_M, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 29).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 2, 29).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 4, 29).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 5, 31).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2016, 6, 30).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2017, 12, 29).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_monthly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_M, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 1, 29).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 2, 29).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 4, 29).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2016, 5, 31).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2016, 6, 30).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-1], datetime(2017, 12, 29).toordinal())

        # test lower frequency data
        self.assertRaises(
            ValueError,
            convert,
            convert(self.ts_ord, new_freq=FREQ_Y),
            new_freq=FREQ_M,
        )

        # timestamp conversion goes here

    def test_convquarterly_period_start(self):
        """
        This function tests converting timeseries to quarterly data that have
        starting period data.

        """

        # conv_quarterly(ts, new_freq):
        ts = self.ts_ord.clone()
        ts.end_of_period = False

        # conv_quarterly with defaults
        ts1 = convert(ts, new_freq=FREQ_Q)

        self.assertEqual(ts1.frequency, FREQ_Q)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 4, 1).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 7, 1).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 10, 3).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2017, 1, 2).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2017, 4, 3).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2017, 7, 3).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_quarterly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_Q, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 4, 1).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 7, 1).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 10, 3).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2017, 1, 2).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2017, 4, 3).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2017, 7, 3).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2018, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_quarterly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_Q, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 4, 1).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 7, 1).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 10, 3).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2017, 1, 2).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2017, 4, 3).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2017, 7, 3).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 1).toordinal())

        # unresolved design decision
        # with monthly data
        # ts1 = convert(
        #    convert(ts, new_freq=FREQ_M, include_partial=False),
        #    new_freq=FREQ_Q,
        #    include_partial=False)

        # self.assertEqual(ts1.dseries[0], datetime(2016, 1, 1).toordinal())
        # self.assertEqual(ts1.dseries[1], datetime(2016, 4, 1).toordinal())
        # self.assertEqual(ts1.dseries[2], datetime(2016, 7, 1).toordinal())
        # self.assertEqual(ts1.dseries[3], datetime(2016, 10, 3).toordinal())
        # self.assertEqual(ts1.dseries[4], datetime(2017, 1, 2).toordinal())
        # self.assertEqual(ts1.dseries[5], datetime(2017, 4, 3).toordinal())
        # self.assertEqual(ts1.dseries[6], datetime(2017, 7, 3).toordinal())

        # ending values
        # self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 1).toordinal())

        # test lower frequency data
        self.assertRaises(
            ValueError,
            convert,
            convert(self.ts_ord, new_freq=FREQ_Y),
            new_freq=FREQ_Q,
        )

        # timestamp conversion goes here

    def test_convquarterly_period_end(self):
        """
        This function tests converting timeseries to quarterly data that have
        ending period data.

        """

        ts = self.ts_ord.clone()

        ts1 = convert(ts, new_freq=FREQ_Q)

        self.assertEqual(ts1.frequency, FREQ_Q)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 6, 30).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 9, 30).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2017, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2017, 6, 30).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2017, 12, 29).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_quarterly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_Q, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 6, 30).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 9, 30).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2017, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2017, 6, 30).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-2], datetime(2017, 12, 29).toordinal())
        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_quarterly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_Q, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2016, 6, 30).toordinal())
        self.assertEqual(ts1.dseries[3], datetime(2016, 9, 30).toordinal())
        self.assertEqual(ts1.dseries[4], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[5], datetime(2017, 3, 31).toordinal())
        self.assertEqual(ts1.dseries[6], datetime(2017, 6, 30).toordinal())

        # ending values
        self.assertEqual(ts1.dseries[-1], datetime(2017, 12, 29).toordinal())

        # resolve design decision
        # ts1 = convert(
        #    convert(ts, new_freq=FREQ_M, include_partial=False),
        #    new_freq=FREQ_Q,
        #    include_partial=False)

        # self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        # self.assertEqual(ts1.dseries[1], datetime(2016, 3, 31).toordinal())
        # self.assertEqual(ts1.dseries[2], datetime(2016, 6, 30).toordinal())
        # self.assertEqual(ts1.dseries[3], datetime(2016, 9, 30).toordinal())
        # self.assertEqual(ts1.dseries[4], datetime(2016, 12, 30).toordinal())
        # self.assertEqual(ts1.dseries[5], datetime(2017, 3, 31).toordinal())
        # self.assertEqual(ts1.dseries[6], datetime(2017, 6, 30).toordinal())

        # ending values
        # self.assertEqual(ts1.dseries[-1], datetime(2017, 12, 29).toordinal())

        # test lower frequency data
        # test lower frequency data
        self.assertRaises(
            ValueError,
            convert,
            convert(self.ts_ord, new_freq=FREQ_Y),
            new_freq=FREQ_Q,
        )

        # timestamp conversion goes here

    def test_convyearly_period_start(self):
        """
        This function tests converting timeseries to yearly data that have
        starting period data.

        """

        ts = self.ts_ord.clone()

        # conv_yearly with defaults
        ts1 = convert(ts, new_freq=FREQ_Y)

        self.assertEqual(ts1.frequency, FREQ_Y)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_yearly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_Y, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_yearly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_Y, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        # resolve design decision
        # with monthly data
        # ts1 = convert(
        #    convert(ts, new_freq=FREQ_M, include_partial=False),
        #    new_freq=FREQ_Y,
        #    include_partial=False)

        # self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        # self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        # self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        # timestamp conversion goes here

    def test_convyearly_period_end(self):
        """
        This function tests converting timeseries to yearly data that have
        ending period data.

        """

        ts = self.ts_ord.clone()

        # conv_yearly with defaults
        ts1 = convert(ts, new_freq=FREQ_Y)

        self.assertEqual(ts1.frequency, FREQ_Y)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_yearly with include_partial=True
        ts1 = convert(ts, new_freq=FREQ_Y, include_partial=True)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        self.assertEqual(ts1.dseries[-1], datetime(2018, 1, 15).toordinal())

        # conv_yearly with include_partial=False
        ts1 = convert(ts, new_freq=FREQ_Y, include_partial=False)

        self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        # resolve design decision
        # with monthly data
        # ts1 = convert(
        #    convert(ts, new_freq=FREQ_M, include_partial=False),
        #    new_freq=FREQ_Y,
        #    include_partial=False)

        # self.assertEqual(ts1.dseries[0], datetime(2015, 12, 31).toordinal())
        # self.assertEqual(ts1.dseries[1], datetime(2016, 12, 30).toordinal())
        # self.assertEqual(ts1.dseries[2], datetime(2017, 12, 29).toordinal())

        # timestamp conversion goes here

    def test_convminutes_period_start(self):
        """
        This function tests converting timeseries to minute data that have
        starting period data.

        """

        ts = self.ts_seconds.clone()
        ts.end_of_period = False

        ts1 = convert(ts, new_freq=FREQ_MIN)

        self.assertEqual(
            ts1.dseries[0], datetime(2016, 1, 1, 0, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[1], datetime(2016, 1, 1, 0, 1, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[2], datetime(2016, 1, 1, 0, 2, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[3], datetime(2016, 1, 1, 0, 3, 0).timestamp()
        )

    @unittest.skip
    def test_convminutes_period_end(self):
        """
        This function tests converting timeseries to minute data that have
        ending period data.

        Needs design decision.
        """

        ts = self.ts_seconds.clone()

        ts1 = convert(ts, new_freq=FREQ_MIN)

        self.assertEqual(
            ts1.dseries[0], datetime(2016, 1, 1, 0, 0, 59).timestamp()
        )
        self.assertEqual(
            ts1.dseries[1], datetime(2016, 1, 1, 0, 1, 59).timestamp()
        )
        self.assertEqual(
            ts1.dseries[2], datetime(2016, 1, 1, 0, 2, 59).timestamp()
        )
        self.assertEqual(
            ts1.dseries[3], datetime(2016, 1, 1, 0, 3, 59).timestamp()
        )
        self.assertEqual(
            ts1.dseries[4], datetime(2016, 1, 1, 0, 4, 59).timestamp()
        )
        self.assertEqual(
            ts1.dseries[5], datetime(2016, 1, 1, 0, 5, 59).timestamp()
        )

    def test_convhours_period_start(self):
        """
        This function tests conversion to hours.

        Currently there is a problem with this. At the moment it a design
        decision needs to be made on how to handle end-of-period conversions.

        """
        ts = self.ts_seconds.clone()
        ts.end_of_period = False
        ts1 = convert(ts, new_freq=FREQ_H)

        self.assertEqual(
            ts1.dseries[0], datetime(2016, 1, 1, 0, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[1], datetime(2016, 1, 1, 1, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[2], datetime(2016, 1, 1, 2, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[3], datetime(2016, 1, 1, 3, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[4], datetime(2016, 1, 1, 4, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[5], datetime(2016, 1, 1, 5, 0, 0).timestamp()
        )

    @unittest.skip
    def test_convhours_period_end(self):
        """
        This function tests conversion to hours.

        Currently there is a problem with this. At the moment a design
        decision needs to be made on how to handle end-of-period conversions.

        """

        ts = self.ts_seconds.clone()
        ts.end_of_period = False
        ts1 = convert(ts, new_freq=FREQ_H)

        ts = self.ts_seconds.clone()
        ts.end_of_period = True
        ts1 = convert(ts, new_freq=FREQ_H)

        self.assertEqual(
            ts1.dseries[0], datetime(2016, 1, 1, 0, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[1], datetime(2016, 1, 1, 1, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[2], datetime(2016, 1, 1, 2, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[3], datetime(2016, 1, 1, 3, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[4], datetime(2016, 1, 1, 4, 0, 0).timestamp()
        )
        self.assertEqual(
            ts1.dseries[5], datetime(2016, 1, 1, 5, 0, 0).timestamp()
        )

    @unittest.skip
    def test_conv_days(self):
        """
        This function converts timestamp data to a daily frequency with
        ordinal dates.

        Currently, there is a problem with this.
        """

        ts = self.ts_seconds.clone()

        ts.end_of_period = False

        ts1 = convert(ts, new_freq=FREQ_D)

        self.assertEqual(
            ts1.dseries[0], datetime(2016, 1, 1, 0, 0, 0).toordinal()
        )
        self.assertEqual(
            ts1.dseries[1], datetime(2016, 1, 2, 0, 0, 0).toordinal()
        )
        self.assertEqual(
            ts1.dseries[2], datetime(2016, 1, 3, 0, 0, 0).toordinal()
        )
        self.assertEqual(
            ts1.dseries[3], datetime(2016, 1, 4, 0, 0, 0).toordinal()
        )

        ts = self.ts_seconds.clone()
        ts.end_of_period = True

        ts1 = convert(ts, new_freq=FREQ_D)

        self.assertEqual(
            ts1.dseries[0], datetime(2016, 1, 1, 0, 0, 0).toordinal()
        )
        self.assertEqual(
            ts1.dseries[1], datetime(2016, 1, 2, 0, 0, 0).toordinal()
        )
        self.assertEqual(
            ts1.dseries[2], datetime(2016, 1, 3, 0, 0, 0).toordinal()
        )
        self.assertEqual(
            ts1.dseries[3], datetime(2016, 1, 4, 0, 0, 0).toordinal()
        )


if __name__ == "__main__":
    unittest.main()



"""
This module implements a class derived from dicts for working with timeseries.

"""

from copy import deepcopy
import json

from .tsslist import TssList
from .timeseries import Timeseries


class TssDict(dict):
    """
    This class is a way of handling some of the routine tasks for groups
    of timeseries.

    Assumption:
        This is a dict of timeseries that are keyed by tickers, etc. Or, it
        could be a dict of keys that hold lists of timeseries with some
        commonality.

    Usage:
        tssdict = TssDict(values=None)

    values can be a dict, list, using the key from each timeseries as the
    dict key.


    """

    def __init__(self, values=None):

        dict.__init__(self)  # only did this to satisfy pylint

        if isinstance(values, dict):
            for key, values in values.items():
                self[key] = values
        elif isinstance(values, list):
            for i, ts_tmp in enumerate(values):
                ts_tmp = values[i]
                self[ts_tmp.key] = ts_tmp
        else:
            # nothing to do.
            pass

    def min_date(self):
        """
        Returns the earliest date as a tuple(datetime, key in the group).
        """
        min_date = None
        min_key = None

        for key, values in self.items():

            if isinstance(values, Timeseries):
                date = values.start_date("datetime")
                if min_date is not None:
                    if date < min_date:
                        min_date = date
                        min_key = key
                else:
                    min_date = date
                    min_key = key

            elif isinstance(values, list):
                if min_date is not None:
                    date = values.min_date()
                    if date < min_date:
                        min_date = date
                        min_key = key
                else:
                    min_date = TssList(values).min_date()
                    min_key = key

            elif isinstance(values, TssDict):
                if min_date is not None:
                    date = values.min_date()
                    if date < min_date:
                        min_date = date
                        min_key = key
                else:
                    min_date, _ = values.min_date()
                    min_key = key

            else:
                # what is it?
                raise ValueError("Unsupported values in dict")

        return (min_date, min_key)

    def max_date(self):
        """
        Returns the latest date, key in the group

        If more than one has the same max date, simply one of them is
        returned.
        """
        max_date = None
        max_key = None

        for key, values in self.items():

            if isinstance(values, Timeseries):
                date = values.end_date("datetime")
                if max_date is not None:
                    date = date
                    if date > max_date:
                        max_date = date
                        max_key = key
                else:
                    max_date = date
                    max_key = key

            elif isinstance(values, list):
                if max_date is not None:
                    date = values.max_date()
                    if date > max_date:
                        max_date = date
                        max_key = key
                else:
                    max_date = TssList(values).max_date()
                    max_key = key

            elif isinstance(values, TssDict):
                if max_date is not None:
                    date = values.max_date()
                    if date > max_date:
                        max_date = date
                        max_key = key
                else:
                    max_date, _ = values.max_date()
                    max_key = key

            else:
                # what is it?
                raise ValueError("Unsupported values in dict")

        return (max_date, max_key)

    def longest_ts(self):
        """
        This function returns item with the longest timeseries.

        """
        max_length = 0
        max_key = None

        for key, values in self.items():

            if isinstance(values, Timeseries):
                length = values.tseries.shape[0]
                if length > max_length:
                    max_length = length
                    max_key = key

            elif isinstance(values, list):
                length = max([ts.tseries.shape[0] for ts in values])
                if length > max_length:
                    max_length = length
                    max_key = key

            elif isinstance(values, TssDict):
                length, _ = values.longest_ts()
                if length > max_length:
                    max_length = length
                    max_key = key

            else:
                # what is it?
                raise ValueError("Unsupported values in dict")

        return (max_length, max_key)

    def shortest_ts(self):
        """
        This function returns item with the shortest timeseries.

        """
        min_length = 0
        min_key = None

        for key, values in self.items():

            if isinstance(values, Timeseries):
                length = values.tseries.shape[0]
                if length < min_length:
                    min_length = length
                    min_key = key

            elif isinstance(values, list):
                length = min([ts.tseries.shape[0] for ts in values])
                if length < min_length:
                    min_length = length
                    min_key = key

            elif isinstance(values, TssDict):
                length, _ = values.longest_ts()
                if length < min_length:
                    min_length = length
                    min_key = key

            else:
                # what is it?
                raise ValueError("Unsupported values in dict")

        return (min_length, min_key)

    def get_values(self, date, keys=None, notify=False):
        """
        This function finds the values as the date. If keys come in as a list
        the order of the values can be controlled or to limit the timeseries
        selected.

        The point of notify is to not fail gracefully if necessary.
        """

        if keys is None:
            keys = self.keys()

        all_values = []

        for key in keys:

            tmp = self[key]

            if isinstance(tmp, TssList):
                all_values.append(tmp.get_values(date, notify=notify))

            elif isinstance(tmp, list):

                all_values.append(TssList(tmp).get_values(date, notify=notify))

            elif isinstance(tmp, Timeseries):

                try:
                    all_values.append(tmp.tseries[tmp.row_no(date)])
                except ValueError:
                    if notify:
                        raise ValueError(
                            "ts %s does not have a value on %s" % (key, date)
                        )
                    else:
                        all_values.append(None)

            else:

                # thinking about doing TssDict but not sure about passing up
                # the values and keys back up.
                pass

        return (tuple(all_values), tuple(keys))

    def combine(self, keys=None, discard=True, pad=None):
        """
        This function combines all timeseries into one. The point of keys is
        that you have the ability to control the order of the columns.

        Usage:
            combine(self, keys=None, pad=None)

        returns ts, keys

        """

        def iter_combine(ts1, item, discard=discard, pad=pad):
            """This function combines an item with an existing timeseries. """
            if isinstance(item, TssList):

                if ts1 is None:
                    ts1 = item.combine(discard=discard, pad=pad)
                else:
                    ts1.combine(item, discard=discard, pad=pad)

            elif isinstance(item, list):

                if ts1 is None:
                    ts1 = TssList(item).combine(discard=discard, pad=pad)
                else:
                    ts1.combine(item, discard=discard, pad=pad)

            elif isinstance(item, Timeseries):

                if ts1 is None:
                    ts1 = item.clone()
                else:
                    ts1 = ts1.combine(item, discard=discard, pad=pad)

            elif isinstance(item, TssDict):

                if ts1 is None:
                    ts1, _ = item.combine(discard=discard, pad=pad)
                else:
                    ts1.combine(
                        item.combine(discard=discard, pad=pad),
                        discard=discard,
                        pad=pad,
                    )

            else:
                raise ValueError("Unsupported type in for \n%s" % (item))

            return ts1

        if keys is None:
            keys = self.keys()

        if len(keys) == 0:
            return None

        ts1 = None
        for key in keys:
            ts1 = iter_combine(ts1, self[key], discard=discard, pad=pad)

        return ts1, tuple(keys)

    def clone(self):
        """
        Returns a new copy of the object.
        """

        return deepcopy(self)

    def to_dict(self, dt_fmt="str", data_list=True):
        """
        This function outputs the entirety of the object as a dict with
        the timeseries components as a dict as well.

        This enables building JSON formatted files from objects that include
        TssDict objects.

        Usage:
            self.to_dict(dt_fmt='str', data_list=True)
        """
        outdict = {}

        for key, ts in self.items():
            outdict[key] = ts.to_dict(dt_fmt=dt_fmt, data_list=data_list)

        return outdict

    def from_dict(self, tssdict):
        """
        This function loads from a dict.
        """
        self.clear()

        for key, value in tssdict.items():
            self[key] = Timeseries().from_dict(value)

        return self

    def to_json(self, indent=2, dt_fmt="str", data_list=True):
        """
        This function returns the timeseries dict in JSON format.

        Usage:
            self.to_json(indent=2, dt_fmt='str', data_list=True)

        indent: indenting in the JSON output
        dt_fmt: formatting of the dates. Look at help for
                    Timeseries.to_dict
        data_list: Whether data uses a dict for dates as keys or
                   simply a list.
                   Default is for a list. Otherwise, sorting the
                   timeseries in the list would be required.

        """
        return json.dumps(
            self.to_dict(dt_fmt=dt_fmt, data_list=data_list), indent=indent
        )

    def from_json(self, json_str):
        """
        This function loads a JSON string and applies it to the object.
        """
        self.clear()

        tss_tmp = json.loads(json_str)

        if isinstance(tss_tmp, dict):

            for key, value in tss_tmp.items():
                self[key] = Timeseries().from_dict(value)

        else:
            ValueError("Incoming JSON string does not start with a dict.")

        return self



"""
This module implements the class Timeseries

"""
import datetime as dt
from copy import deepcopy
import json
import numpy as np

from .constants import TS_ORDINAL, FREQ_D, FREQ_M
from .constants import TS_TIMESTAMP
from .constants import FREQ_DAYTYPES, FREQ_IDAYTYPES

from .freq_conversions import convert

from .tsproto import TsProto

FMT_DATE = "%Y-%m-%d"
FMT_IDATE = "%Y-%m-%d %H:%M:%S"


class Timeseries(TsProto):
    """
    This class holds timeseries data. Dates and values are kept in
    separate numpy arrays.

    Usage:

        ts = Timeseries(
            frequency=FREQ_D,   defaults to daily
            dseries=None,       a numpy array with ordinal / timestamps
            tseries=None,       a numpy array of values
            end_of_period=True, did these values take place during the period?
            key='',             an optional key for naming the timeseries
            columns=None        optional columns for naming columns of the
                                values
        )
    """

    def __init__(
        self,
        frequency=FREQ_D,
        end_of_period=None,
        key="",
        columns=None,
        **kwargs
    ):

        TsProto.__init__(self)

        self.frequency = frequency
        self.key = key

        if "tseries" in kwargs:
            self.tseries = kwargs["tseries"]

        if "dseries" in kwargs:
            self.dseries = kwargs["dseries"]

        if end_of_period is not None:
            self.end_of_period = end_of_period

        if key:
            self.key = key

        if columns is not None:
            self.columns = columns

    def series_direction(self):
        """
        if a lower row is a lower date, then 1 for ascending
        if a lower row is a higher date then -1 for descending
        """
        if len(self.dseries) > 1:
            if self.dseries[0] < self.dseries[-1]:
                return 1
            else:
                return -1
        else:
            return 0

    def start_date(self, fmt=None):
        """
        This function returns the starting date of the timeseries in its
        native value, timestamp or ordinal.

        If fmt is 'str' returns in string format
        If fmt is 'datetime' returns as datetime object

        CHANGE: this now returns
            a date object if ordinal
            a datetime if timestamp

        Note: look at consolidating return date formats:
            end_date, get_datetime
        """
        if self.series_direction() == 1:
            row_no = 0
        else:
            row_no = -1

        if fmt is None:
            return self.dseries[row_no]
        elif fmt == "str":
            return self.fmt_date(
                self.dseries[row_no], dt_type=self.get_date_series_type()
            )
        elif fmt == "datetime":
            return self.get_datetime(self.dseries[row_no])
        else:
            raise ValueError("Invalid date format fmt: %s" % fmt)

    def end_date(self, fmt=None):
        """
        This funtcion returns the ending date of the timeseries in its native
        value, timestamp or ordinal.

        If fmt is 'str' returns in string format
        If fmt is 'datetime' returns as datetime object
        """
        if self.series_direction() == 1:
            row_no = -1
        else:
            row_no = 0

        if fmt is None:
            return self.dseries[row_no]
        elif fmt == "str":
            return self.fmt_date(
                self.dseries[row_no], dt_type=self.get_date_series_type()
            )
        elif fmt == "datetime":
            return self.get_datetime(self.dseries[row_no])
        else:
            raise ValueError("Invalid date format fmt: %s" % fmt)

    def get_datetime(self, date):
        """
        This function returns a date as a datetime object.
        This takes into account the type of date stored in dseries.

        Usage:
            self.get_datetime(date)
        """
        date_series_type = self.get_date_series_type()
        if date_series_type == TS_ORDINAL:
            return dt.date.fromordinal(date)
        elif date_series_type == TS_TIMESTAMP:
            return dt.datetime.fromtimestamp(date)
        else:
            raise ValueError(
                "Unknown, dateseries type. %s" % (date_series_type)
            )

    def to_dict(self, dt_fmt=None, data_list=False):
        """
        Returns the time series as a dict with the date as the key.

        Usage:
            self.to_dict(dt_fmt=None, data_list=False)

        This has been reworked to include all fields of the timeseries
        rather than just dates and times, so header informtion is now included.

        Also, for flexibility, the date can be formatted in various ways:
            dt_fmt=None       -- native format depending on frequency
                                 but converted to string
            dt_fmt='datetime' -- datetime objects
            dt_fmt='str'      -- converts dates to string using
                                 FMT_DATE or FMT_IDATE, depending on the
                                 timeseries type
        The boolean, data_list, signals whether dates should be used
        as keys in a dict for the values, or whether the dates and values
        are output as a list.

        This matters because some operations are necessary to target
        specific dates, but it does not preserve order. Or, if data_list
        is True, then the combination of dates and values are output as
        a list and order is maintained.

        """
        new_dict = {}
        new_dict["header"] = self.header()

        if dt_fmt is None:
            # native format, but converted to a string
            dseries = [str(date) for date in self.dseries.tolist()]
        elif dt_fmt == "datetime":
            dseries = self.datetime_series()
        elif dt_fmt == "str":
            dseries = self.date_string_series()
        else:
            raise ValueError("Unsupported dt_fmt: %s" % (dt_fmt))

        tseries = self.tseries.tolist()
        if data_list:
            data = [(dseries[i], tseries[i]) for i in range(len(self.tseries))]
        else:
            data = dict(
                [(dseries[i], tseries[i]) for i in range(len(self.tseries))]
            )

        new_dict["data"] = data

        return new_dict

    def from_dict(self, ts_dict):
        """
        This function accepts a dict and assembles a timeseries.

        Usage:
            self.from_dict(ts_dict)

        The dict must be:
            {
                "header": {
                     items
                 },
                "data": [
                    [date1, value1],
                    [date2, value2]
                ]
            }

        Or,
            {
                "header": {
                    items
                },
                "data": {
                    date1: values,
                    date2: values,
                }
            }

        The date formatting uses the same rules as creation in self.to_dict.

        If the data portion is a dict rather than a list, the timeseries is
        sorted in reverse data order by default.

        It loads in-place, but also returns self for use with chaining.
        """
        # header
        header = ts_dict["header"]
        for key, value in header.items():
            self.__setattr__(key, value)

        data = ts_dict["data"]

        do_sort = False
        if isinstance(data, dict):
            do_sort = True
            data = list(data.items())

        # dseries
        if self.get_date_series_type() == TS_ORDINAL:
            fmt = FMT_DATE
            self.dseries = [
                dt.datetime.strptime(item[0], fmt).toordinal() for item in data
            ]
        elif self.get_date_series_type() == TS_TIMESTAMP:
            fmt = FMT_IDATE
            self.dseries = [
                dt.datetime.strptime(item[0], fmt).timestamp() for item in data
            ]
        else:
            raise ValueError("undefined frequency: %s" % self.frequency)

        # tseries
        self.tseries = [item[1] for item in data]

        self.make_arrays()

        if do_sort:
            self.sort_by_date(reverse=True, force=True)

        return self

    def to_list(self):
        """ Returns the timeseries as a list. """

        return [
            (str(self.dseries[i]), deepcopy(self.tseries[i]))
            for i in range(self.shape()[0])
        ]

    def to_json(self, indent=2, dt_fmt="str", data_list=True):
        """
        This function returns the timeseries in JSON format.

        Usage:
            self.as_json(indent=2, dt_fmt=str, data_list=True)

        dt_fmt options are the same as for to_dict

        """
        return json.dumps(
            self.to_dict(dt_fmt, data_list=data_list), indent=indent
        )

    def from_json(self, json_str):
        """
        This function loads a string in JSON format to inhabit a Timeseries
        object.

        Usage:
            self.from_json(json_str)

        It looks for a format similar to the following:
        {
            "header": {
                "end_of_period": True,
                "key": "",
                "columns": [],
                "frequency": "d"
            },
            "data": [
                ["2015-12-31", 0.0],
                ["2016-01-01", 1.0],
                ["2016-01-02", 2.0],
                ["2016-01-03", 3.0],
                ["2016-01-04", 4.0]
            ]
        }

        """
        tmpdict = json.loads(json_str)

        self.from_dict(tmpdict)

        return self

    def extend(self, ts, overlay=True):
        """
        This function combines a timeseries to another, taking into account the
        possibility of overlap.

        This assumes that the frequency is the same.

        This function is chiefly envisioned to extend a timeseries with
        additional dates.

        Usage:
            self.extend(ts, overlay=True)

        If overlay is True then the incoming timeseries will overlay
        any values that are duplicated.

        NOTE: add a check to prevent unwanted overlays
        """
        series_dir = self.series_direction()
        if series_dir == -1:
            reverse = True
        else:
            reverse = False

        self_dict = self.to_dict()["data"]
        ts_dict = ts.to_dict()["data"]

        for tdate in ts_dict.keys():
            if tdate in self_dict.keys():
                if overlay:
                    self_dict[tdate] = ts_dict[tdate]
                else:
                    raise ValueError(
                        "Duplicate dates, overlay parameter is False: %s"
                        % (tdate)
                    )
            else:
                self_dict[tdate] = ts_dict[tdate]

        self.dseries = []
        self.tseries = []
        for date, value in sorted(self_dict.items(), reverse=reverse):
            self.dseries.append(date)
            self.tseries.append(value)

        self.make_arrays()

    def add(self, ts, match=True):
        """
        Adds two timeseries together.

        if match is True:
            means there should be a one to one corresponding date in each time
            series.  If not raise error.
        else:
            means that timeseries with sporadic or missing dates can be added

        Note: this does not evaluate whether both timeseries have the same
                number of columns. It will fail if they do not.

        Returns the timeseries. Not in-place.
        """
        self._column_checks(self, ts)

        series_dir = self.series_direction()
        if series_dir == -1:
            reverse = True
        else:
            reverse = False

        self_ts = self.clone()
        tmp_ts = ts.clone()
        shape = tmp_ts.shape()
        if match:

            if len(self_ts.tseries) != len(tmp_ts.tseries):
                raise ValueError("Timeseries do not have the same length.")

            if self.if_dseries_match(ts) is False:
                raise ValueError("Dateseries do not have the same dates.")

            #   ok
            self_ts.tseries += tmp_ts.tseries
            return self_ts

        else:
            #   Dates do not have to match up. return an aglomeration of both
            dates = self_ts.to_dict()["data"]
            dates1 = tmp_ts.to_dict()["data"]

            if len(shape) > 1:
                # convert back to np arrays for the addition
                for key, value in list(dates.items()):
                    dates[key] = np.array(value)
                for key, value in list(dates1.items()):
                    dates1[key] = np.array(value)

            for pdate, value in dates1.items():
                if isinstance(value, float):
                    dates.setdefault(pdate, 0)
                else:
                    dates.setdefault(pdate, np.zeros(shape))
                dates[pdate] += value

            self_ts.dseries = []
            self_ts.tseries = []
            for date, value in sorted(dates.items(), reverse=reverse):
                self_ts.dseries.append(date)
                self_ts.tseries.append(value)

            self_ts.make_arrays()

            return self_ts

    def replace(self, ts, match=True):
        """
        This function replaces values where the dates match an incoming
        timeseries. So if the incoming date on the timeseries matches, the
        value in the current timeseries will be replaced by the incoming
        timeseries.

        Usage:
            self.replace(ts, match=True)

        If match is False, the incoming timseries may have dates not found in
        the self timeseries.

        Returns the modified timeseries. Not in place.

        """
        series_dir = self.series_direction()
        if series_dir == -1:
            reverse = True
        else:
            reverse = False

        self_ts = self.clone()
        tmp_ts = ts.clone()

        self_dates = self_ts.to_dict()["data"]
        ts_dates = tmp_ts.to_dict()["data"]

        if match:
            for date, value in ts_dates.items():
                if date in self_dates:
                    self_dates[date] = value

        else:
            for date, value in ts_dates.items():
                self_dates[date] = value

        self_ts.dseries = []
        self_ts.tseries = []
        for date, value in sorted(self_dates.items(), reverse=reverse):
            self_ts.dseries.append(date)
            self_ts.tseries.append(value)

        self_ts.make_arrays()

        return self_ts

    def combine(self, tss, discard=True, pad=None):
        """
        This function combines timeseries into a single array. Combining in
        this case means accumulating additional columns of information.

        Truncation takes place at the end of rows. So if the timeseries is
        sorted from latest dates to earliest dates, the older values would be
        removed.

        Usage:
            self.combine(tss, discard=True, pad=None)

        If discard:
            Will truncate all timeseries lengths down to the shortest
            timeseries.

        if discard is False
            an error will be raised if the all the lengths do not match

            unless:
                if pad is not None:
                    the shorter timeseries will be padded with the value pad.

        Returns the new ts.
        """
        series_dir = self.series_direction()
        if isinstance(tss, Timeseries):
            # A single, make into a list of timeseries
            tss = [tss]

        tss = [self.clone()] + [ts_tmp.clone() for ts_tmp in tss]
        for tmp_ts in tss:
            if len(tmp_ts.tseries.shape) == 1:
                tmp_ts.tseries = tmp_ts.tseries.reshape((-1, 1))

        # check lengths
        if discard is False:
            if pad is None:
                length = len(self.tseries)
                for tmp_ts in tss:
                    if length != len(tmp_ts.tseries):
                        raise ValueError("Lengths are not the same.")
            else:
                # get max length
                max_lens = [tmp_ts.tseries.shape[0] for tmp_ts in tss]
                max_len = max(max_lens)

                ts_ref = tss[max_lens.index(max_len)]
                ts_ref_len = ts_ref.tseries.shape[0]

                for i, tmp_ts in enumerate(tss):
                    tmp_ts = tss[i].clone()
                    ts_len = tmp_ts.tseries.shape[0]

                    # ready to pad
                    if ts_ref_len - ts_len > 0:
                        col_count = tmp_ts.tseries.shape[1]
                        pad_values = (
                            np.ones((ts_ref_len - ts_len, col_count)) * pad
                        )

                        # dates added to the end
                        if ts_len < max_len:
                            tmp_ts.tseries = np.append(
                                tmp_ts.tseries, pad_values
                            ).reshape((-1, col_count))
                            tmp_ts.dseries = np.append(
                                tmp_ts.dseries, ts_ref.dseries[ts_len:]
                            )

                        # put it back
                        tss[i] = tmp_ts
        else:

            length = min([len(tmp_ts.tseries) for tmp_ts in tss])
            for tmp_ts in tss:
                tmp_ts.trunc(finish=length)

        # all timeseries same length
        base_ts = tss[0]
        base_ts.tseries = np.hstack([ts_tmp.tseries for ts_tmp in tss])

        # force a sort
        if series_dir == 1:
            base_ts.sort_by_date(force=True)
        else:
            base_ts.sort_by_date(reverse=True, force=True)
        return base_ts

    def date_string_series(self, dt_fmt=None):
        """
        This function returns a list of the dates in the timeseries as
        strings.

        Usage:
            self.date_string_series(dt_fmt=None)

        dt_fmt is a datetime mask to alter the default formatting.
        """
        dt_type = self.get_date_series_type()
        if dt_fmt is None:
            # use default
            if dt_type == TS_ORDINAL:
                dt_fmt = FMT_DATE
            else:
                dt_fmt = FMT_IDATE

        return [self.fmt_date(date, dt_type, dt_fmt) for date in self.dseries]

    def sort_by_date(self, reverse=False, force=False):
        """
        This function converts a timeseries to either date order or reverse
        date order.

        Usage:
            sort_by_date(self, reverse=False, force=False)

        If reverse is True, then order will be newest to oldest.
        If force is False, the assumption is made that comparing the first
        and last date will determine the current order of the timeseries. That
        would mean that unnecessary sorting can be avoided. Also, if the order
        needs to be reversed, the sort is changed via the less expensive
        reverse function.

        If dates and values are in no particular order, with force=True, the
        actual sort takes place.

        This function changes the data in-place.
        """
        if force:
            dseries = []
            tseries = []
            for date, values in sorted(
                self.to_dict()["data"].items(), reverse=reverse
            ):
                dseries.append(date)
                tseries.append(values)
            self.dseries = dseries
            self.tseries = tseries
            self.make_arrays()

        else:
            if reverse is False and self.series_direction() == 1:
                # unnecessary
                pass
            elif reverse is True and self.series_direction() == -1:
                # unnecessary
                pass
            else:
                self.reverse()

    def reverse(self):
        """
        This function does in-place reversal of the timeseries and dateseries.
        """
        if len(self.tseries.shape) == 1:
            self.tseries = self.tseries[::-1]
        else:
            self.tseries = np.flipud(self.tseries)

        self.dseries = self.dseries[::-1]

    def convert(self, new_freq, include_partial=True, **kwargs):
        """
        This function returns the timeseries converted to another frequency,
        such as daily to monthly.

        Usage:
            convert(new_freq, include_partial=True, **kwargs)

        The only kwarg is
            weekday=<some value>

        This is used when converting to weekly data. The weekday number
        corresponds to the the datetime.weekday() function.
        """
        if new_freq not in FREQ_IDAYTYPES + FREQ_DAYTYPES:
            raise ValueError("Invalid new frequency: %s" % new_freq)

        return convert(
            self, new_freq, include_partial=include_partial, **kwargs
        )

    def get_diffs(self):
        """
        This function gets the differences between values from date to date
        in the timeseries.
        """
        series_dir = self.series_direction()

        tmp_ts = self.clone()
        if series_dir == 1:
            tmp_ts.reverse()

        tmp_ts.tseries = tmp_ts.tseries[:-1] - tmp_ts.tseries[1:]
        tmp_ts.dseries = tmp_ts.dseries[:-1]

        if series_dir == 1:
            tmp_ts.reverse()

        return tmp_ts

    def get_pcdiffs(self):
        """
        This function gets the percent differences between values in the
        timeseries.

        No provision for dividing by zero here.
        """
        series_dir = self.series_direction()

        tmp_ts = self.clone()

        if series_dir == 1:
            tmp_ts.reverse()

        tmp_ts.tseries = (
            (tmp_ts.tseries[:-1] / tmp_ts.tseries[1:]) - 1.0
        ) * 100.0
        tmp_ts.dseries = tmp_ts.dseries[:-1]

        if series_dir == 1:
            tmp_ts.reverse()

        return tmp_ts

    def items(self, fmt=None):
        """This function returns the date series and the time series as if it
        is in one list. The term items used to suggest the iteration of dicts
        where items are the key, value combination.

        if fmt == 'str':
            the dates are output as strings
        """
        if fmt == "str":
            return [
                (date, value.tolist())
                for date, value in zip(self.date_string_series(), self.tseries)
            ]
        else:
            return [
                (date, value)
                for date, value in zip(self.datetime_series(), self.tseries)
            ]

    def trunc(self, start=None, finish=None, new=False):
        """
        This function truncates in place, typically.

        truncate from (start:finish)
        remember start is lowest number, latest date

        This truncation works on the basis of slicing, so
        finish is not inclusive.

        Usage:
            self.trunc(start=None, finish=None, new=False)

        If new is True, the timeseries will not be modified in place. Rather
        a new timeseries will be returned instead.
        """
        if new:
            tmp_ts = self.clone()
        else:
            tmp_ts = self

        if start and finish:
            tmp_ts.tseries = tmp_ts.tseries[start:finish]
            tmp_ts.dseries = tmp_ts.dseries[start:finish]
        elif start:
            tmp_ts.tseries = tmp_ts.tseries[start:]
            tmp_ts.dseries = tmp_ts.dseries[start:]
        elif finish:
            tmp_ts.tseries = tmp_ts.tseries[:finish]
            tmp_ts.dseries = tmp_ts.dseries[:finish]

        if new:
            return tmp_ts

    def truncdate(self, start=None, finish=None, new=False):
        """

        This function truncates in place on the basis of dates.

        Usage:
            self.truncdate(start=None, finish=None, new=False)

        start and finish are dates, input as either datetime or the actual
        internal format of the dseries (ordinals or timestamps).

        If the dates are not actually in the list, the starting date will
        be the next viable date after the start date requested. If the finish
        date is not available, the previous date from the finish date will be
        the last.

        If new is True, the timeseries will not be modified in place. Rather
        a new timeseries will be returned instead.
        """
        if isinstance(start, list) or isinstance(start, tuple):
            start, finish = start

        series_dir = self.series_direction()
        if series_dir == -1:
            self.reverse()

        if new:
            tmp_ts = self.clone()

        if start and finish:

            start = min(start, finish)
            finish = max(start, finish)

            start_row = self.row_no(start, closest=1)
            finish_row = self.row_no(finish, closest=-1)

            if new:
                tmp_ts = self[start_row : finish_row + 1]
            else:
                self.trunc(start=start_row, finish=finish_row + 1)

        elif start:
            # test if date exists
            start_row = self.row_no(start, closest=1)

            if new:
                tmp_ts = self[start_row:]
            else:
                self.trunc(start=start_row)

        elif finish:
            finish_row = self.row_no(finish, closest=-1)

            if new:
                tmp_ts = self[: finish_row + 1]
            else:
                self.trunc(finish=finish_row + 1)

        if series_dir == -1:
            self.reverse()
            if new:
                tmp_ts.reverse()

        if new:
            return tmp_ts

    def row_no(self, rowdate, closest=0, no_error=False):
        """
        Shows the row in the timeseries

        Usage:
            ts.row_no(rowdate=<datetime>)
            ts.row_no(rowdate=<date as either ordinal or timestamp>)

        Returns an error if the date is not found in the index

        if closest is invoked:
            closest = 1
                find the closest date after the rowdate
            closest = -1
                find the closest date before the rowdate

        If no_error
            returns -1 instead of raising an error if the date was
            outside of the timeseries.
        """
        row_error = -1

        series_dir = self.series_direction()

        if isinstance(rowdate, dt.datetime) or isinstance(rowdate, dt.date):
            rdate = self.date_native(rowdate)
        else:
            # assume it is appropriate
            rdate = rowdate

        if closest not in [-1, 0, 1]:
            raise ValueError("Invalid closest value: %s" % (closest))

        if closest == 0:

            selected = np.argwhere(self.dseries == rdate)

            if selected.shape[0] == 0:
                if no_error:
                    row_no = row_error
                else:
                    raise ValueError(
                        "%s not found in %s timeseries" % (rowdate, self.key)
                    )
            else:
                row_no = selected[0][0]

        elif closest == -1:
            selected = np.argwhere(self.dseries <= rdate)

            if selected.shape[0] == 0:
                if no_error:
                    row_no = row_error
                else:
                    raise ValueError(
                        "%s not found in %s timeseries" % (rowdate, self.key)
                    )

            if series_dir == 1:
                row_no = selected.max()
            else:
                row_no = selected.min()

        else:
            selected = np.argwhere(self.dseries >= rdate)

            if selected.shape[0] == 0:
                if no_error:
                    row_no = row_error
                else:
                    raise ValueError(
                        "%s not found in %s timeseries" % (rowdate, self.key)
                    )

            if series_dir == 1:
                row_no = selected.min()
            else:
                row_no = selected.max()

        return row_no

    def datetime_series(self):
        """
        This function returns the dateseries converted to a series of
        datetime objects.
        """
        if self.get_date_series_type() == TS_ORDINAL:
            return [dt.date.fromordinal(int(i)) for i in self.dseries]
        elif self.get_date_series_type() == TS_TIMESTAMP:
            return [dt.datetime.fromtimestamp(int(i)) for i in self.dseries]
        else:
            raise ValueError("timeseries must have a defined frequency")

    @staticmethod
    def fmt_date(numericdate, dt_type, dt_fmt=None):
        """
        This static method accepts a date and converts it to
        the format used in the timeseries.
        """
        if dt_type == TS_ORDINAL:
            if dt_fmt is None:
                dt_fmt = FMT_DATE
            return dt.date.fromordinal(int(numericdate)).strftime(dt_fmt)
        elif dt_type == TS_TIMESTAMP:
            if dt_fmt is None:
                dt_fmt = FMT_IDATE
            return dt.datetime.fromtimestamp(numericdate).strftime(dt_fmt)
        else:
            raise ValueError("Unknown dt_type: %s" % dt_type)

    def __repr__(self):
        """
        This function returns a representation of the class.
        """
        output = "\n".join(
            [
                "<Timeseries>",
                "key: " + self.key,
                "columns: %s" % self.columns,
                "frequency: %s" % self.frequency,
                "daterange: %s" % str(self.daterange("str")),
                "end-of-period: %s" % self.end_of_period,
                "shape: %s" % str(self.shape()),
            ]
        )

        return output

    def set_zeros(self, fmt=None, new=False):
        """
        This function converts an existing timeseries to zeros using the same
        shape as the existing timeseries.

        It is used as a convenience to create an empty timeseries with a
        specified date range.

        if fmt use as shape

        usage:
            set_zeros(self, fmt=None, new=False)
        """
        if new:
            tmp_ts = self.clone()
        else:
            tmp_ts = self

        if fmt is not None:
            tmp_ts.tseries = np.zeros(fmt)
        else:
            tmp_ts.tseries = np.zeros(tmp_ts.tseries.shape)

        if new:
            return tmp_ts

    def set_ones(self, fmt=None, new=False):
        """
        This function converts an existing timeseries to ones using the same
        shape as the existing timeseries.

        It is used as a convenience to create an empty timeseries with a
        specified date range.

        If fmt use as shape

        usage:
            set_ones(self, fmt=None, new=False)
        """
        if new:
            tmp_ts = self.clone()
        else:
            tmp_ts = self

        if fmt is not None:
            tmp_ts.tseries = np.ones(fmt)
        else:
            tmp_ts.tseries = np.ones(tmp_ts.tseries.shape)

        if new:
            return tmp_ts

    def header(self):
        """
        This function returns a dict of the non-timeseries data.

        Changed:
        Non-timeseries data is determined as an element of __dict__ that is
        not tseries or dseries.

        This enables more descriptive data to be included in the header.
        """
        return dict(
            [
                (key, value)
                for key, value in self.__dict__.items()
                if key not in ["tseries", "dseries"]
            ]
        )

    def date_native(self, date):
        """
        This awkwardly named function returns a date in the native format of
        of the timeseries, namely ordinal or timestamp.
        """
        if isinstance(date, dt.datetime) or isinstance(date, dt.date):
            datetype = self.get_date_series_type()

            if datetype == TS_ORDINAL:
                ndate = date.toordinal()
            else:
                ndate = date.timestamp()

            return ndate
        else:
            raise ValueError("date must be a datetime")

    def daterange(self, fmt=None):
        """
        This function returns the starting and ending dates of the timeseries.

        Usage:

            self.daterange()
                (735963, 735972)

            self.daterange('str')
                ('2015-12-31', '2016-01-09')

            self.daterange('datetime')
                (datetime(2015, 12, 31, 0, 0),
                 datetime.datetime(2016, 1, 9, 0, 0))

        """

        if self.dseries is not None:
            start_date, end_date = (self.dseries.min(), self.dseries.max())
        else:
            return (None, None)

        if fmt == "str":
            dt_type = self.get_date_series_type()
            start_date = self.fmt_date(start_date, dt_type)
            end_date = self.fmt_date(end_date, dt_type)
        elif fmt == "datetime":
            start_date = self.get_datetime(start_date)
            end_date = self.get_datetime(end_date)
        elif fmt is None:
            #   either timestamp or ordinal, pass through as is
            pass
        else:
            raise ValueError("Invalid fmt type: %s" % (fmt))

        return (start_date, end_date)

    def years(self, include_partial=True):
        """
        This function provides a quick way to summarize yearly data.

        It is basically a pass-through to the convert function with more
        decoration of the years.

        Usage:

            years(include_partial=True)

            returns a dict with year as keys
        """
        ts_years = convert(
            self, new_freq=FREQ_M, include_partial=include_partial
        )

        year_dict = {}

        tseries = ts_years.tseries
        dseries = ts_years.dseries

        for i in range(tseries.shape[0]):
            year = ts_years.get_datetime(dseries[i]).year
            year_dict[year] = tseries[i]

        return year_dict

    def months(self, include_partial=True):
        """
        This function provides a quick way to summarize monthly data.

        It is basically a pass-through to the convert function with more
        decoration of the months.

        Usage:

            months(include_partial=True)

            returns a dict with year-month as keys
        """
        ts_months = convert(
            self, new_freq=FREQ_M, include_partial=include_partial
        )

        month_dict = {}

        tseries = ts_months.tseries
        dseries = ts_months.dseries

        for i in range(tseries.shape[0]):
            date = ts_months.get_datetime(dseries[i])
            month = "%s-%02d" % (date.year, date.month)

            month_dict[month] = tseries[i]

        return month_dict

    def closest_date(self, rowdate, closest=1):
        """
        This function is a variation on the self.row_no function. The date
        presumably is not going to be found in the date series. Instead,
        given the parameter of closest, it eithers to either a row to the past
        or the future.

        Usage:
            self.closest_date(rowdate=date)             search forward
            self.closest_date(rowdate=date, closest=1)  search forward
            self.closest_date(rowdate=date, closest=-1) search backwards
        """
        row_no = self.row_no(rowdate=rowdate, closest=closest)

        return self.dseries[row_no]

    def get_duped_dates(self):
        """
        This function pulls dates that are duplicated. This is to be used to
        locate timeseries that are faulty.

        Usage:
            get_duped_dates()

            returns [[odate1, count], [odate2, count]]
        """
        dict_dates = {}

        for odate in self.dseries:
            dict_dates.setdefault(odate, 0)
            dict_dates[odate] += 1

        return [
            [odate, count] for odate, count in dict_dates.items() if count > 1
        ]

    def get_fromDB(self, **kwargs):
        """
        This is just a stub to suggest a viable name for getting data from
        a database.
        """
        pass

    def save_toDB(self, **kwargs):
        """
        This is just a stub to suggest a viable name for saving data to a
        database.
        """
        pass
